# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc

import tof

Hz = sc.Unit('Hz')
deg = sc.Unit('deg')
rad = sc.Unit('rad')
meter = sc.Unit('m')
sec = sc.Unit('s')
two_pi = 2.0 * rad * sc.constants.pi


def test_angular_speed():
    f = 10.0 * Hz
    chopper = tof.Chopper(
        frequency=f,
        open=0.0 * deg,
        close=10.0 * deg,
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )
    assert chopper.omega == two_pi * f


def test_open_close_times_one_rotation():
    f = 10.0 * Hz
    chopper = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0], unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )

    topen, tclose = chopper.open_close_times(0 * sec)
    # Note that choppers perform one rotation before t=0 to make sure
    # no openings are missed at early times.
    assert sc.identical(topen[0], ((10.0 * deg).to(unit='rad') - two_pi) / (two_pi * f))
    assert sc.identical(
        tclose[0], ((20.0 * deg).to(unit='rad') - two_pi) / (two_pi * f)
    )
    assert sc.identical(topen[1], (10.0 * deg).to(unit='rad') / (two_pi * f))
    assert sc.identical(tclose[1], (20.0 * deg).to(unit='rad') / (two_pi * f))


def test_open_close_times_three_rotations():
    f = 10.0 * Hz
    chopper = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0], unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )

    topen, tclose = chopper.open_close_times(0.21 * sec)
    open0 = (10.0 * deg).to(unit='rad')
    close0 = (20.0 * deg).to(unit='rad')
    assert sc.identical(
        topen,
        sc.concat(
            [open0 - two_pi, open0, open0 + two_pi, open0 + 2 * two_pi], dim='cutout'
        )
        / (two_pi * f),
    )
    assert sc.identical(
        tclose,
        sc.concat(
            [close0 - two_pi, close0, close0 + two_pi, close0 + 2 * two_pi],
            dim='cutout',
        )
        / (two_pi * f),
    )


def test_open_close_angles_scalars_converted_to_arrays():
    f = 10.0 * Hz
    chopper = tof.Chopper(
        frequency=f,
        open=10.0 * deg,
        close=20.0 * deg,
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )
    topen, tclose = chopper.open_close_times(0.0 * sec)
    assert sc.identical(topen[1], (10.0 * deg).to(unit='rad') / (two_pi * f))
    assert sc.identical(tclose[1], (20.0 * deg).to(unit='rad') / (two_pi * f))


def test_phase():
    f = 10.0 * Hz
    chopper1 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0], unit='deg'),
        phase=0.0 * deg,
        distance=10.0 * meter,
        name='chopper1',
    )
    topen1, tclose1 = chopper1.open_close_times(0.0 * sec)
    chopper2 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0], unit='deg'),
        phase=30.0 * deg,
        distance=10.0 * meter,
        name='chopper2',
    )
    topen2, tclose2 = chopper2.open_close_times(0.0 * sec)
    assert sc.allclose(topen2, topen1 + (30.0 * deg).to(unit='rad') / chopper2.omega)
    assert sc.allclose(tclose2, tclose1 + (30.0 * deg).to(unit='rad') / chopper2.omega)


def test_phase_int():
    f = 10.0 * Hz
    op = sc.array(dims=['cutout'], values=[10.0], unit='deg')
    cl = sc.array(dims=['cutout'], values=[20.0], unit='deg')
    d = 10.0 * meter
    chopper1 = tof.Chopper(
        frequency=f,
        open=op,
        close=cl,
        phase=30.0 * deg,
        distance=d,
        name='chopper1',
    )
    chopper2 = tof.Chopper(
        frequency=f,
        open=op,
        close=cl,
        phase=30 * deg,
        distance=d,
        name='chopper2',
    )
    topen1, tclose1 = chopper1.open_close_times(0.0 * sec)
    topen2, tclose2 = chopper2.open_close_times(0.0 * sec)
    assert sc.allclose(topen1, topen2)
    assert sc.allclose(tclose1, tclose2)


def test_frequency_must_be_positive():
    with pytest.raises(ValueError, match="Chopper frequency must be positive"):
        tof.Chopper(
            frequency=-1.0 * Hz,
            open=0.0 * deg,
            close=10.0 * deg,
            phase=0.0 * deg,
            distance=5.0 * meter,
            name='chopper',
        )


def test_open_close_times_counter_rotation():
    f = 10.0 * Hz
    d = 10.0 * meter
    ph = 0.0 * deg
    chopper1 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0, 90.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0, 130.0], unit='deg'),
        phase=ph,
        distance=d,
        name='chopper1',
    )
    chopper2 = tof.Chopper(
        frequency=f,
        open=sc.array(
            dims=['cutout'], values=[360.0 - 130.0, 360.0 - 20.0], unit='deg'
        ),
        close=sc.array(
            dims=['cutout'], values=[360.0 - 90.0, 360.0 - 10.0], unit='deg'
        ),
        phase=ph,
        distance=d,
        direction=tof.AntiClockwise,
        name='chopper2',
    )

    topen1, tclose1 = chopper1.open_close_times(0.0 * sec)
    topen2, tclose2 = chopper2.open_close_times(0.0 * sec)
    assert sc.allclose(topen1, topen2)
    assert sc.allclose(tclose1, tclose2)


def test_open_close_times_counter_rotation_with_phase():
    f = 10.0 * Hz
    chopper1 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[80.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[90.0], unit='deg'),
        phase=0.0 * deg,
        distance=10.0 * meter,
        direction=tof.AntiClockwise,
        name='chopper1',
    )
    topen1, tclose1 = chopper1.open_close_times(0.0 * sec)
    chopper2 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[80.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[90.0], unit='deg'),
        phase=30.0 * deg,
        distance=10.0 * meter,
        direction=tof.AntiClockwise,
        name='chopper2',
    )
    topen2, tclose2 = chopper2.open_close_times(0.0 * sec)
    assert sc.allclose(
        topen2, topen1 + (30.0 * deg).to(unit='rad') / abs(chopper2.omega)
    )
    assert sc.allclose(
        tclose2, tclose1 + (30.0 * deg).to(unit='rad') / abs(chopper2.omega)
    )


def test_open_close_anticlockwise_multiple_rotations():
    chopper = tof.Chopper(
        frequency=10.0 * Hz,
        open=sc.array(dims=["cutout"], values=[10.0, 90.0], unit="deg"),
        close=sc.array(dims=["cutout"], values=[20.0, 130.0], unit="deg"),
        distance=10.0 * meter,
        phase=0 * deg,
        direction=tof.AntiClockwise,
        name='chopper',
    )

    two_rotations_open, two_rotations_close = chopper.open_close_times(0.0 * sec)
    three_rotations_open, three_rotations_close = chopper.open_close_times(0.2 * sec)
    four_rotations_open, four_rotations_close = chopper.open_close_times(0.3 * sec)

    assert len(two_rotations_open) == 4
    assert len(three_rotations_open) == 6
    assert len(four_rotations_open) == 8
    assert len(two_rotations_close) == 4
    assert len(three_rotations_close) == 6
    assert len(four_rotations_close) == 8

    assert sc.allclose(two_rotations_open, three_rotations_open[:-2])
    assert sc.allclose(two_rotations_close, three_rotations_close[:-2])
    assert sc.allclose(three_rotations_open, four_rotations_open[:-2])
    assert sc.allclose(three_rotations_close, four_rotations_close[:-2])
    assert sc.allclose(two_rotations_open, four_rotations_open[:-4])
    assert sc.allclose(two_rotations_close, four_rotations_close[:-4])


def test_bad_direction_raises():
    f = 10.0 * Hz
    op = sc.array(dims=['cutout'], values=[10.0], unit='deg')
    cl = sc.array(dims=['cutout'], values=[20.0], unit='deg')
    d = 10.0 * meter
    ph = 0.0 * deg
    tof.Chopper(
        frequency=f,
        open=op,
        close=cl,
        phase=ph,
        distance=d,
        direction=tof.Clockwise,
        name='chopper',
    )
    tof.Chopper(
        frequency=f,
        open=op,
        close=cl,
        phase=ph,
        distance=d,
        direction=tof.AntiClockwise,
        name='chopper',
    )
    with pytest.raises(
        ValueError, match="Chopper direction must be Clockwise or AntiClockwise"
    ):
        tof.Chopper(
            frequency=f,
            open=op,
            close=cl,
            phase=ph,
            distance=d,
            direction='clockwise',
            name='chopper',
        )
    with pytest.raises(
        ValueError, match="Chopper direction must be Clockwise or AntiClockwise"
    ):
        tof.Chopper(
            frequency=f,
            open=op,
            close=cl,
            phase=ph,
            distance=d,
            direction='anti-clockwise',
            name='chopper',
        )
    with pytest.raises(
        ValueError, match="Chopper direction must be Clockwise or AntiClockwise"
    ):
        tof.Chopper(
            frequency=f,
            open=op,
            close=cl,
            phase=ph,
            distance=d,
            direction=1,
            name='chopper',
        )


def test_chopper_create_from_centers_and_widths():
    f = 10.0 * Hz
    centers = sc.array(dims=['cutout'], values=[15.0, 46.0], unit='deg')
    widths = sc.array(dims=['cutout'], values=[10.0, 16.0], unit='deg')
    chopper = tof.Chopper(
        frequency=f,
        centers=centers,
        widths=widths,
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )
    assert sc.allclose(chopper.open, centers - widths / 2)
    assert sc.allclose(chopper.close, centers + widths / 2)

    expected = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0, 38.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0, 54.0], unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )
    assert sc.allclose(chopper.open, expected.open)
    assert sc.allclose(chopper.close, expected.close)


def test_chopper_create_raises_when_both_edges_and_centers_are_supplied():
    with pytest.raises(
        ValueError, match="Either open/close or centers/widths must be provided"
    ):
        tof.Chopper(
            frequency=10.0 * Hz,
            open=10.0 * deg,
            close=20.0 * deg,
            centers=15.0 * deg,
            widths=10.0 * deg,
            phase=0.0 * deg,
            distance=5.0 * meter,
            name='chopper',
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

from functools import reduce
from itertools import chain
from types import MappingProxyType
from typing import Dict, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import scipp as sc
from matplotlib.collections import LineCollection

from .chopper import Chopper, ChopperReading
from .detector import Detector, DetectorReading
from .reading import ReadingData, ReadingField
from .source import Source, SourceParameters
from .utils import Plot


def _make_reading_data(component, dim, is_chopper=False):
    visible = {}
    blocked = {} if is_chopper else None
    keep_dim = (set(component.dims) - {'pulse'}).pop()
    for name, da in sc.collapse(component, keep=keep_dim).items():
        one_mask = ~reduce(lambda a, b: a | b, da.masks.values())
        vsel = da[one_mask]
        visible[name] = sc.DataArray(data=vsel.data, coords={dim: vsel.coords[dim]})
        if is_chopper:
            bsel = da[da.masks['blocked_by_me']]
            blocked[name] = sc.DataArray(data=bsel.data, coords={dim: bsel.coords[dim]})
    return ReadingField(
        visible=ReadingData(data=sc.DataGroup(visible), dim=dim),
        blocked=(
            ReadingData(data=sc.DataGroup(blocked), dim=dim) if is_chopper else None
        ),
    )


def _add_rays(
    ax: plt.Axes,
    toas: sc.Variable,
    birth_times: sc.Variable,
    distances: sc.Variable,
    cbar: bool = True,
    cmap: str = 'gist_rainbow_r',
    wavelengths: Optional[sc.Variable] = None,
    wmin: Optional[sc.Variable] = None,
    wmax: Optional[sc.Variable] = None,
):
    x0 = birth_times.values.reshape(-1, 1)
    x1 = toas.values.reshape(-1, 1)
    y0 = np.zeros(x0.size).reshape(-1, 1)
    y1 = distances.values.reshape(-1, 1)
    segments = np.concatenate(
        (
            np.concatenate((x0, y0), axis=1).reshape(-1, 1, 2),
            np.concatenate((x1, y1), axis=1).reshape(-1, 1, 2),
        ),
        axis=1,
    )
    coll = LineCollection(segments)
    if wavelengths is not None:
        coll.set_cmap(plt.colormaps[cmap])
        coll.set_array(wavelengths.values)
        coll.set_norm(plt.Normalize(wmin.value, wmax.value))
        if cbar:
            cb = plt.colorbar(coll, ax=ax)
            cb.ax.yaxis.set_label_coords(-0.9, 0.5)
            cb.set_label('Wavelength (Å)')
    else:
        coll.set_color('lightgray')
    ax.add_collection(coll)


class Result:
    """
    Result of a simulation.

    Parameters
    ----------
    source:
        The source of neutrons.
    choppers:
        The choppers in the model.
    detectors:
        The detectors in the model.
    """

    def __init__(
        self,
        source: Source,
        choppers: Dict[str, Chopper],
        detectors: Dict[str, Detector],
    ):
        self._source = source.as_readonly()
        self._masks = {}
        self._arrival_times = {}
        self._choppers = {}
        fields = {
            'toas': 'toa',
            'wavelengths': 'wavelength',
            'birth_times': 'time',
            'speeds': 'speed',
        }
        for name, chopper in choppers.items():
            self._masks[name] = chopper['visible_mask']
            self._arrival_times[name] = chopper['data'].coords['toa']
            self._choppers[name] = ChopperReading(
                distance=chopper['distance'],
                name=chopper['name'],
                frequency=chopper['frequency'],
                open=chopper['open'],
                close=chopper['close'],
                phase=chopper['phase'],
                open_times=chopper['open_times'],
                close_times=chopper['close_times'],
                data=chopper['data'],
                **{
                    key: _make_reading_data(chopper['data'], dim=dim, is_chopper=True)
                    for key, dim in fields.items()
                },
            )

        self._detectors = {}
        for name, det in detectors.items():
            self._masks[name] = det['visible_mask']
            self._arrival_times[name] = det['data'].coords['toa']
            self._detectors[name] = DetectorReading(
                distance=det['distance'],
                name=det['name'],
                data=det['data'],
                **{
                    key: _make_reading_data(det['data'], dim=dim)
                    for key, dim in fields.items()
                },
            )

        self._choppers = MappingProxyType(self._choppers)
        self._detectors = MappingProxyType(self._detectors)
        self._masks = MappingProxyType(self._masks)
        self._arrival_times = MappingProxyType(self._arrival_times)

    @property
    def choppers(self) -> MappingProxyType[str, ChopperReading]:
        """The choppers in the model."""
        return self._choppers

    @property
    def detectors(self) -> MappingProxyType[str, DetectorReading]:
        """The detectors in the model."""
        return self._detectors

    @property
    def source(self) -> SourceParameters:
        """The source of neutrons."""
        return self._source

    def __iter__(self):
        return chain(self._choppers, self._detectors)

    def __getitem__(self, name: str) -> Union[ChopperReading, DetectorReading]:
        if name not in self:
            raise KeyError(f"No component with name {name} was found.")
        return self._choppers[name] if name in self._choppers else self._detectors[name]

    def _plot_visible_rays(
        self,
        max_rays: int,
        pulse_index: int,
        furthest_detector: DetectorReading,
        ax: plt.Axes,
        cbar: bool,
        wmin: sc.Variable,
        wmax: sc.Variable,
        cmap: str,
    ):
        da = furthest_detector.data['pulse', pulse_index]
        visible = da[~da.masks['blocked_by_others']]
        toas = visible.coords['toa']
        if (max_rays <= 0) or (len(toas) == 0):
            return
        if (max_rays is not None) and (len(toas) > max_rays):
            inds = np.random.choice(len(toas), size=max_rays, replace=False)
        else:
            inds = slice(None)
        birth_times = visible.coords['time'][inds]
        wavelengths = visible.coords['wavelength'][inds]
        distances = furthest_detector.distance.broadcast(sizes=birth_times.sizes)
        _add_rays(
            ax=ax,
            toas=toas[inds],
            birth_times=birth_times,
            distances=distances,
            cbar=cbar,
            wavelengths=wavelengths,
            wmin=wmin,
            wmax=wmax,
            cmap=cmap,
        )

    def _plot_blocked_rays(
        self,
        blocked_rays: int,
        pulse_index: int,
        furthest_detector: DetectorReading,
        ax: plt.Axes,
    ):
        slc = ('pulse', pulse_index)
        inv_mask = ~self._masks[furthest_detector.name][slc]
        nrays = int(inv_mask.sum())
        if (blocked_rays <= 0) or (nrays == 0):
            return
        if nrays > blocked_rays:
            inds = np.random.choice(nrays, size=blocked_rays, replace=False)
        else:
            inds = slice(None)
        birth_times = self._source.data.coords['time'][slc][inv_mask][inds]

        components = sorted(
            chain(self._choppers.values(), [furthest_detector]),
            key=lambda c: c.distance.value,
        )
        dim = 'component'
        toas = sc.concat(
            [
                self._arrival_times[comp.name][slc][inv_mask][inds]
                for comp in components
            ],
            dim=dim,
        )
        distances = sc.concat(
            [comp.distance.broadcast(sizes=birth_times.sizes) for comp in components],
            dim=dim,
        )
        masks = sc.concat(
            [sc.ones(sizes=birth_times.sizes, dtype=bool)]
            + [self._masks[comp.name][slc][inv_mask][inds] for comp in components],
            dim=dim,
        )

        diff = sc.abs(masks[dim, 1:].to(dtype=int) - masks[dim, :-1].to(dtype=int))
        diff.unit = ''
        _add_rays(
            ax=ax,
            toas=(toas * diff).max(dim=dim),
            birth_times=birth_times,
            distances=(distances * diff).max(dim=dim),
        )

    def _plot_pulse(self, pulse_index: int, ax: plt.Axes):
        time_coord = self.source.data.coords['time']['pulse', pulse_index]
        tmin = time_coord.min().value
        ax.plot(
            [tmin, time_coord.max().value],
            [0, 0],
            color="gray",
            lw=3,
        )
        ax.text(tmin, 0, "Pulse", ha="left", va="top", color="gray")

    def plot(
        self,
        max_rays: int = 1000,
        blocked_rays: int = 0,
        figsize: Optional[Tuple[float, float]] = None,
        ax: Optional[plt.Axes] = None,
        cbar: bool = True,
        cmap: str = 'gist_rainbow_r',
    ) -> Plot:
        """
        Plot the time-distance diagram for the instrument, including the rays of
        neutrons that make it to the furthest detector.
        As plotting many lines can be slow, the number of rays to plot can be
        limited by setting ``max_rays``.
        In addition, it is possible to also plot the rays that are blocked by
        choppers along the flight path by setting ``blocked_rays > 0``.

        Parameters
        ----------
        max_rays:
            Maximum number of rays to plot.
        blocked_rays:
            Number of blocked rays to plot.
        figsize:
            Figure size.
        ax:
            Axes to plot on.
        cbar:
            Show a colorbar for the wavelength if ``True``.
        cmap:
            Colormap to use for the wavelength colorbar.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = ax.get_figure()
        furthest_detector = max(self._detectors.values(), key=lambda d: d.distance)
        wavelengths = sc.DataArray(
            data=furthest_detector.data.coords['wavelength'],
            masks=furthest_detector.data.masks,
        )
        for i in range(self._source.data.sizes['pulse']):
            self._plot_blocked_rays(
                blocked_rays=blocked_rays,
                pulse_index=i,
                furthest_detector=furthest_detector,
                ax=ax,
            )
            self._plot_visible_rays(
                max_rays=max_rays,
                pulse_index=i,
                furthest_detector=furthest_detector,
                ax=ax,
                cbar=cbar and (i == 0),
                wmin=wavelengths.min(),
                wmax=wavelengths.max(),
                cmap=cmap,
            )
            self._plot_pulse(pulse_index=i, ax=ax)

        det_data = furthest_detector.toas.visible.data
        if sum(da.sum().value for da in det_data.values()) > 0:
            times = (da.coords['toa'].max() for da in det_data.values())
        else:
            times = (ch.close_times.max() for ch in self._choppers.values())
        toa_max = reduce(max, times).value
        dx = 0.05 * toa_max
        # Plot choppers
        for ch in self._choppers.values():
            x0 = ch.open_times.values
            x1 = ch.close_times.values
            x = np.empty(3 * x0.size, dtype=x0.dtype)
            x[0::3] = x0
            x[1::3] = 0.5 * (x0 + x1)
            x[2::3] = x1
            x = np.concatenate(
                ([[0]] if x[0] > 0 else [x[0:1]])
                + [x]
                + ([[toa_max + dx]] if x[-1] < toa_max else [])
            )
            y = np.full_like(x, ch.distance.value)
            y[2::3] = None
            inds = np.argsort(x)
            ax.plot(x[inds], y[inds], color="k")
            ax.text(
                toa_max, ch.distance.value, ch.name, ha="right", va="bottom", color="k"
            )

        # Plot detectors
        for det in self._detectors.values():
            ax.plot([0, toa_max], [det.distance.value] * 2, color="gray", lw=3)
            ax.text(
                0, det.distance.value, det.name, ha="left", va="bottom", color="gray"
            )

        ax.set_xlabel("Time-of-flight (us)")
        ax.set_ylabel("Distance (m)")
        ax.set_xlim(0 - dx, toa_max + dx)
        if figsize is None:
            inches = fig.get_size_inches()
            fig.set_size_inches(
                (
                    min(inches[0] * furthest_detector.data.sizes['pulse'], 12.0),
                    inches[1],
                )
            )
        fig.tight_layout()
        return Plot(fig=fig, ax=ax)

    def __repr__(self) -> str:
        source_sizes = self._source.data.sizes
        other_dim = (set(source_sizes) - {'pulse'}).pop()
        out = (
            f"Result:\n  Source: {source_sizes['pulse']} pulses, "
            f"{source_sizes[other_dim]} neutrons per pulse.\n  Choppers:\n"
        )
        for name, ch in self._choppers.items():
            out += f"    {name}: {ch.toas._repr_string_body()}\n"
        out += "  Detectors:\n"
        for name, det in self._detectors.items():
            out += f"    {name}: {det.toas._repr_string_body()}\n"
        return out

    def __str__(self) -> str:
        return self.__repr__()

    def to_nxevent_data(self, key: str) -> sc.DataArray:
        """
        Convert a component reading to event data that resembles event data found in a
        NeXus file.

        Parameters
        ----------
        key:
            Name of the component.
        """
        start = sc.datetime("2024-01-01T12:00:00.000000")
        period = sc.reciprocal(self.source.frequency)
        raw_data = self[key].data.flatten(to='event')
        # Select only the neutrons that make it to the detector
        event_data = raw_data[~raw_data.masks['blocked_by_others']].copy()
        dt = period.to(unit=event_data.coords['toa'].unit)
        event_time_zero = (dt * (event_data.coords['toa'] // dt)).to(dtype=int) + start
        event_data.coords['event_time_zero'] = event_time_zero
        event_data.coords['event_time_offset'] = event_data.coords.pop(
            'toa'
        ) % period.to(unit=dt.unit)
        return (
            event_data.drop_coords(['tof', 'speed', 'time', 'wavelength'])
            .group('event_time_zero')
            .rename_dims(event_time_zero='pulse')
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)

import os
import sys

src = os.path.abspath('../src')
os.environ['PYTHONPATH'] = src
sys.path.insert(0, src)

# -- Project information -----------------------------------------------------

project = 'tof'
copyright = '2023, Scipp contributors'
author = 'Neil Vaytet'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'nbsphinx',
    'sphinx_copybutton',
]

autodoc_typehints = 'description'

autosummary_generate = True

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = True
napoleon_use_rtype = False
napoleon_preprocess_types = True
napoleon_type_aliases = {
    # objects without namespace: scipp
    "DataArray": "~scipp.DataArray",
    "Dataset": "~scipp.Dataset",
    "Variable": "~scipp.Variable",
    # objects without namespace: numpy
    "ndarray": "~numpy.ndarray",
}
typehints_defaults = 'comma'
typehints_use_rtype = False


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# source_suffix = ['.rst', '.md']
source_suffix = '.rst'
html_sourcelink_suffix = ''  # Avoid .ipynb.txt extensions in sources

# The master toctree document.
master_doc = 'index'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme'
html_theme_options = {
    'logo_only': True,
    "repository_url": "https://github.com/scipp/tof",
    "repository_branch": "main",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "show_toc_level": 2,
}
html_logo = "_static/logo.svg"
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import uuid
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Tuple

import scipp as sc

from .deprecation import deprecated
from .reading import ComponentReading, ReadingField
from .utils import two_pi


class Direction(Enum):
    CLOCKWISE = 1
    ANTICLOCKWISE = 2


Clockwise = Direction.CLOCKWISE
AntiClockwise = Direction.ANTICLOCKWISE


class Chopper:
    """
    A chopper is a rotating device with cutouts that blocks the beam at certain times.

    Parameters
    ----------
    frequency:
        The frequency of the chopper. Must be positive.
    distance:
        The distance from the source to the chopper.
    name:
        The name of the chopper.
    phase:
        The phase of the chopper. Because the phase offset implemented as a time delay
        on real beamline choppers, it is applied in the opposite direction
        to the chopper rotation direction. For example, if the chopper rotates
        clockwise, a phase of 10 degrees will shift all window angles by 10 degrees
        in the anticlockwise direction, which will result in the windows opening later.
    open:
        The opening angles of the chopper cutouts.
    close:
        The closing angles of the chopper cutouts.
    centers:
        The centers of the chopper cutouts.
    widths:
        The widths of the chopper cutouts.

    Notes
    -----
    Either `open` and `close` or `centers` and `widths` must be provided, but not both.
    """

    def __init__(
        self,
        *,
        frequency: sc.Variable,
        distance: sc.Variable,
        name: str,
        phase: Optional[sc.Variable] = None,
        open: Optional[sc.Variable] = None,
        close: Optional[sc.Variable] = None,
        centers: Optional[sc.Variable] = None,
        widths: Optional[sc.Variable] = None,
        direction: Direction = Clockwise,
    ):
        if frequency <= (0.0 * frequency.unit):
            raise ValueError(f"Chopper frequency must be positive, got {frequency:c}.")
        self.frequency = frequency.to(dtype=float, copy=False)
        if direction not in (Clockwise, AntiClockwise):
            raise ValueError(
                "Chopper direction must be Clockwise or AntiClockwise"
                f", got {direction}."
            )
        if phase is None:
            phase = sc.scalar(0.0, unit='deg')
        self.direction = direction
        # Check that either open/close or centers/widths are provided, but not both
        if tuple(x for x in (open, close, centers, widths) if x is not None) not in (
            (open, close),
            (centers, widths),
        ):
            raise ValueError(
                "Either open/close or centers/widths must be provided, got"
                f" open={open}, close={close}, centers={centers}, widths={widths}."
            )
        if open is None:
            half_width = widths * 0.5
            open = centers - half_width
            close = centers + half_width

        self.open = (open if open.dims else open.flatten(to='cutout')).to(
            dtype=float, copy=False
        )
        self.close = (close if close.dims else close.flatten(to='cutout')).to(
            dtype=float, copy=False
        )
        self.distance = distance.to(dtype=float, copy=False)
        self.phase = phase.to(dtype=float, copy=False)
        self.name = name
        super().__init__()

    @property
    def omega(self) -> sc.Variable:
        """
        The angular velocity of the chopper.
        """
        return two_pi * self.frequency

    def open_close_times(
        self, time_limit: Optional[sc.Variable] = None, unit: Optional[str] = None
    ) -> Tuple[sc.Variable, sc.Variable]:
        """
        The times at which the chopper opens and closes.

        Parameters
        ----------
        time_limit:
            Determines how many rotations the chopper needs to perform to reach the time
            limit. If not specified, the chopper will perform a single rotation.
        unit:
            The unit of the returned times. If not specified, the unit of `time_limit`
            is used.
        """
        if time_limit is None:
            time_limit = sc.scalar(0.0, unit='us')
        if unit is None:
            unit = time_limit.unit
        nrot = max(int(sc.ceil((time_limit * self.frequency).to(unit='')).value), 1)
        # Start at -1 to catch early openings in case the phase or opening angles are
        # large
        phases = sc.arange(uuid.uuid4().hex, -1, nrot) * two_pi + self.phase.to(
            unit='rad'
        )

        open_times = self.open.to(unit='rad', copy=False)
        close_times = self.close.to(unit='rad', copy=False)
        # If the chopper is rotating anti-clockwise, we mirror the openings because the
        # first cutout will be the last to open.
        if self.direction == AntiClockwise:
            open_times, close_times = (
                sc.array(
                    dims=close_times.dims,
                    values=(two_pi - close_times).values[::-1],
                    unit=close_times.unit,
                ),
                sc.array(
                    dims=open_times.dims,
                    values=(two_pi - open_times).values[::-1],
                    unit=open_times.unit,
                ),
            )
        # Note that the order is important here: we need (phases + open/close) to get
        # the correct dimension order when we flatten.
        open_times = (phases + open_times).flatten(to=self.open.dim)
        close_times = (phases + close_times).flatten(to=self.close.dim)
        open_times /= self.omega
        close_times /= self.omega
        return (
            open_times.to(unit=unit, copy=False),
            close_times.to(unit=unit, copy=False),
        )

    def __repr__(self) -> str:
        return (
            f"Chopper(name={self.name}, distance={self.distance:c}, "
            f"frequency={self.frequency:c}, phase={self.phase:c}, "
            f"direction={self.direction.name}, cutouts={len(self.open)})"
        )

    def as_dict(self):
        return {
            'frequency': self.frequency,
            'open': self.open,
            'close': self.close,
            'distance': self.distance,
            'phase': self.phase,
            'name': self.name,
            'direction': self.direction,
        }


@dataclass(frozen=True)
class ChopperReading(ComponentReading):
    """
    Read-only container for the neutrons that reach the chopper.
    """

    distance: sc.Variable
    name: str
    frequency: sc.Variable
    open: sc.Variable
    close: sc.Variable
    phase: sc.Variable
    open_times: sc.Variable
    close_times: sc.Variable
    data: sc.DataArray
    toas: ReadingField
    wavelengths: ReadingField
    birth_times: ReadingField
    speeds: ReadingField

    def __repr__(self) -> str:
        out = f"ChopperReading: '{self.name}'\n"
        out += f"  distance: {self.distance:c}\n"
        out += f"  frequency: {self.frequency:c}\n"
        out += f"  phase: {self.phase:c}\n"
        out += f"  cutouts: {len(self.open)}\n"
        out += "\n".join(
            f"  {key}: {getattr(self, key)}"
            for key in ('toas', 'wavelengths', 'birth_times', 'speeds')
        )
        return out

    def __str__(self) -> str:
        return self.__repr__()

    @property
    @deprecated("Use 'toas' instead.")
    def tofs(self) -> ReadingField:
        return self.toas



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from dataclasses import dataclass
from typing import Optional, Union

import matplotlib.pyplot as plt
import plopp as pp
import scipp as sc

from .utils import Plot


@dataclass(frozen=True)
class ReadingData:
    """
    A data object contains the data (visible or blocked) for a component data
    (e.g. time-of-flight or wavelengths).

    Parameters
    ----------
    data:
        The data to hold.
    dim:
        The dimension label of the data.
    """

    data: sc.DataGroup
    dim: str

    def __len__(self) -> int:
        """
        The number of pulses in the data.
        """
        return len(self.data)

    def __getitem__(self, val: Union[int, slice]):
        """
        Get the data for a single pulse or a range of pulses.

        Parameters
        ----------
        val:
            The index or slice of the pulse(s) to get.
        """
        if isinstance(val, int):
            val = slice(val, val + 1)
        # Convert the slice to a list of indices, which can then be used to create
        # DataGroup keys in the form 'pulse:0', 'pulse:1', etc.
        inds = range(len(self))[val]
        return self.__class__(
            data=sc.DataGroup(
                {f"pulse:{ind}": self.data[f"pulse:{ind}"] for ind in inds}
            ),
            dim=self.dim,
        )

    def plot(self, bins: Union[int, sc.Variable] = 300, **kwargs):
        """
        Plot the neutrons that reach the component as a histogram.

        Parameters
        ----------
        bins:
            The bins to use for histogramming the neutrons.
        """
        return pp.plot(
            {
                key: da.hist({self.dim: bins})
                for key, da in self.data.items()
                if da.size > 0
            },
            **kwargs,
        )

    def __repr__(self) -> str:
        out = f"ReadingData(dim='{self.dim}')\n"
        for name, da in self.data.items():
            out += f"  {name}: events={len(da)}"
            if len(da) > 0:
                coord = da.coords[self.dim]
                out += f", min={coord.min():c}, max={coord.max():c})"
            out += "\n"
        return out

    def __str__(self) -> str:
        return self.__repr__()


def _field_to_string(field: ReadingData) -> str:
    if isinstance(field.data, sc.DataArray):
        return str(len(field))
    data = field.data
    out = [str(len(data["pulse:0"]))]
    npulses = len(data)
    if npulses > 2:
        out.append("...")
    if npulses > 1:
        out.append(str(len(data[f"pulse:{npulses - 1}"])))
    return "[" + ", ".join(out) + "]"


@dataclass(frozen=True)
class ReadingField:
    """
    Contains the data for the visible neutrons of a given field.
    Possible fields are ``toas``, ``wavelengths``, ``birth_times``, and ``speeds``.
    In the case of a :class:`Chopper`, this also contains the data for the blocked
    neutrons.

    Parameters
    ----------
    visible:
        The visible neutrons (those that are not blocked by the component).
    blocked:
        The blocked neutrons (those that are blocked by the component).
    """

    visible: ReadingData
    blocked: Optional[ReadingData] = None

    @property
    def data(self) -> sc.DataGroup:
        """
        The neutrons that reach the component, split up into those that are blocked by
        the component and those that are not.
        """
        out = {"visible": self.visible.data}
        if self.blocked is not None:
            out["blocked"] = self.blocked.data
        return sc.DataGroup(out)

    def __getitem__(self, val):
        return self.__class__(
            visible=self.visible[val],
            blocked=self.blocked[val] if self.blocked is not None else None,
        )

    def _repr_string_body(self) -> str:
        out = f"visible={_field_to_string(self.visible)}"
        if self.blocked is not None:
            out += f", blocked={_field_to_string(self.blocked)}"
        return out

    def __repr__(self) -> str:
        return f"ReadingData(dim='{self.visible.dim}', {self._repr_string_body()})"

    def plot(self, bins: Union[int, sc.Variable] = 300, **kwargs):
        """
        Plot the data for the neutrons that reach the component, split up into those
        that are blocked by the component and those that are not.

        Parameters
        ----------
        bins:
            The bins to use for histogramming the neutrons.
        """
        if self.blocked is None:
            return self.visible.plot(bins=bins, **kwargs)
        visible = self.visible.data
        blocked = self.blocked.data
        if isinstance(visible, sc.DataArray):
            visible = sc.DataGroup({"onepulse": visible})
            blocked = sc.DataGroup({"onepulse": blocked})
        dim = self.visible.dim
        to_plot = {}
        colors = {}
        edges = bins
        for i, p in enumerate(visible):
            if isinstance(bins, int):
                edges = sc.linspace(
                    dim=dim,
                    start=min(
                        visible[p].coords[dim].min(), blocked[p].coords[dim].min()
                    ).value,
                    stop=max(
                        visible[p].coords[dim].max(), blocked[p].coords[dim].max()
                    ).value,
                    num=bins,
                    unit=visible[p].coords[dim].unit,
                )
            vk = f"visible-{p}"
            bk = f"blocked-{p}"
            to_plot.update(
                {vk: visible[p].hist({dim: edges}), bk: blocked[p].hist({dim: edges})}
            )
            colors.update({vk: f"C{i}", bk: "gray"})
        out = pp.plot(
            to_plot,
            **{**{"color": colors}, **kwargs},
        )
        return out


class ComponentReading:
    """
    Data reading for a component placed in the beam path. The reading will have a
    record of the arrival times and wavelengths of the neutrons that passed through it.
    """

    def plot(self, bins: int = 300) -> Plot:
        """
        Plot both the toa and wavelength data side by side.

        Parameters
        ----------
        bins:
            Number of bins to use for histogramming the neutrons.
        """
        fig, ax = plt.subplots(1, 2)
        self.toas.plot(bins=bins, ax=ax[0])
        self.wavelengths.plot(bins=bins, ax=ax[1])
        size = fig.get_size_inches()
        fig.set_size_inches(size[0] * 2, size[1])
        fig.tight_layout()
        return Plot(fig=fig, ax=ax)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from dataclasses import dataclass

import scipp as sc

from .deprecation import deprecated
from .reading import ComponentReading, ReadingField


class Detector:
    """
    A detector component does not block any neutrons, it sees all neutrons passing
    through it.

    Parameters
    ----------
    distance:
        The distance from the source to the detector.
    name:
        The name of the detector.
    """

    def __init__(self, distance: sc.Variable, name: str):
        self.distance = distance.to(dtype=float, copy=False)
        self.name = name

    def __repr__(self) -> str:
        return f"Detector(name={self.name}, distance={self.distance:c})"

    def as_dict(self):
        return {'distance': self.distance, 'name': self.name}


@dataclass(frozen=True)
class DetectorReading(ComponentReading):
    """
    Read-only container for the neutrons that reach the detector.
    """

    distance: sc.Variable
    name: str
    data: sc.DataArray
    toas: ReadingField
    wavelengths: ReadingField
    birth_times: ReadingField
    speeds: ReadingField

    def __repr__(self) -> str:
        out = f"DetectorReading: '{self.name}'\n"
        out += f"  distance: {self.distance:c}\n"
        out += "\n".join(
            f"  {key}: {getattr(self, key)}"
            for key in ('toas', 'wavelengths', 'birth_times', 'speeds')
        )
        return out

    def __str__(self) -> str:
        return self.__repr__()

    @property
    @deprecated("Use 'toas' instead.")
    def tofs(self) -> ReadingField:
        return self.toas



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


import numpy as np
import pytest
import scipp as sc

import tof


def test_ess_pulse():
    source = tof.Source(facility='ess', neutrons=100_000)
    # Check that the time distribution is low on edges and high in the middle
    times = source.data['pulse', 0].hist(time=300)
    mean = times.mean()
    assert (times[0] < 0.5 * mean).value
    assert (times[-1] < 0.5 * mean).value
    assert (times[150] > 1.5 * mean).value
    # Check that there are more neutrons at low wavelengths
    wavs = source.data['pulse', 0].hist(wavelength=300)
    assert (wavs[:150].sum() > 1.5 * wavs[150:].sum()).value


def test_creation_from_neutrons():
    birth_times = sc.array(dims=['event'], values=[1000.0, 1500.0, 2000.0], unit='us')
    wavelengths = sc.array(dims=['event'], values=[1.0, 5.0, 10.0], unit='angstrom')
    source = tof.Source.from_neutrons(
        birth_times=birth_times,
        wavelengths=wavelengths,
    )
    assert source.neutrons == 3
    assert sc.identical(source.data['pulse', 0].coords['time'], birth_times)
    assert sc.identical(source.data['pulse', 0].coords['wavelength'], wavelengths)


def test_creation_from_distribution_flat():
    time = sc.array(dims=['time'], values=[1.0, 3.0], unit='ms')
    p_time = sc.DataArray(
        data=sc.ones(sizes=time.sizes),
        coords={'time': time},
    )
    wavelength = sc.array(dims=['wavelength'], values=[1.0, 10.0], unit='angstrom')
    p_wav = sc.DataArray(
        data=sc.ones(sizes=wavelength.sizes),
        coords={'wavelength': wavelength},
    )

    N = 123456
    source = tof.Source.from_distribution(
        neutrons=N,
        p_time=p_time,
        p_wav=p_wav,
    )

    assert source.neutrons == N
    assert source.data.sizes['event'] == N
    h = source.data['pulse', 0].hist(time=10)
    assert sc.allclose(
        h.data,
        sc.full(value=N / 10.0, sizes={'time': 10}, unit='counts'),
        rtol=sc.scalar(0.05),
    )


def test_creation_from_distribution():
    v = np.ones(9) * 0.1
    v[3:6] = 1.0

    p_time = sc.DataArray(
        data=sc.array(dims=['time'], values=v),
        coords={'time': sc.linspace('time', 0.0, 8000.0, len(v), unit='us')},
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0]),
        coords={
            'wavelength': sc.array(
                dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0], unit='angstrom'
            )
        },
    )

    source = tof.Source.from_distribution(neutrons=100_000, p_time=p_time, p_wav=p_wav)
    assert source.neutrons == 100_000
    da = source.data['pulse', 0]
    assert da.hist(
        time=sc.array(dims=['time'], values=[2100.0, 2900.0], unit='us')
    ).data.sum() > sc.scalar(0.0, unit='counts')
    assert da.hist(
        time=sc.array(dims=['time'], values=[5100.0, 5900.0], unit='us')
    ).data.sum() > sc.scalar(0.0, unit='counts')

    left = da.hist(
        time=sc.array(dims=['time'], values=[0.0, 2000.0], unit='us')
    ).data.sum()
    mid = da.hist(
        time=sc.array(dims=['time'], values=[3000.0, 5000.0], unit='us')
    ).data.sum()
    right = da.hist(
        time=sc.array(dims=['time'], values=[6000.0, 8000.0], unit='us')
    ).data.sum()
    rtol = sc.scalar(0.05)
    assert sc.isclose(mid / left, sc.scalar(10.0), rtol=rtol)
    assert sc.isclose(mid / right, sc.scalar(10.0), rtol=rtol)

    # Make sure distribution is monotonically increasing
    locs = np.linspace(1.0, 4.0, 20)
    step = 0.5 * (locs[1] - locs[0])
    for i in range(len(locs) - 2):
        a = da.hist(
            wavelength=sc.array(
                dims=['wavelength'],
                values=[locs[i] - step, locs[i] + step],
                unit='angstrom',
            )
        ).data.sum()
        b = da.hist(
            wavelength=sc.array(
                dims=['wavelength'],
                values=[locs[i + 1] - step, locs[i + 1] + step],
                unit='angstrom',
            )
        ).data.sum()
        assert b > a


def test_non_integer_sampling():
    N = 1_000_000
    source_float = tof.Source(facility='ess', neutrons=N, sampling=1e4)
    source_int = tof.Source(facility='ess', neutrons=N, sampling=10_000)
    assert source_float.neutrons == source_int.neutrons == N
    tedges = sc.linspace('time', 0.0, 5.0e3, 301, unit='us')
    wedges = sc.linspace('wavelength', 0.0, 20.0, 301, unit='angstrom')

    da_f = source_float.data['pulse', 0]
    da_i = source_int.data['pulse', 0]

    a = da_f.hist(time=tedges).data
    b = da_i.hist(time=tedges).data
    c = da_f.hist(wavelength=wedges).data
    d = da_i.hist(wavelength=wedges).data

    assert sc.allclose(a, b, atol=0.1 * a.max())
    assert sc.allclose(c, d, atol=0.1 * c.max())


def test_non_integer_neutrons():
    source = tof.Source(facility='ess', neutrons=1e5)
    assert source.neutrons == 100_000


def test_multiple_pulses_ess():
    source = tof.Source(facility='ess', neutrons=100_000, pulses=3)
    assert source.data.sizes['pulse'] == 3
    assert source.data.sizes['event'] == 100_000
    assert sc.identical(source.frequency, sc.scalar(14.0, unit='Hz'))
    bins = (sc.arange('time', 4) / source.frequency).to(unit='us')
    h = source.data.flatten(to='event').hist(time=bins)
    assert sc.allclose(h.data, sc.full(value=1e5, sizes={'time': 3}, unit='counts'))


def test_multiple_pulses_from_distribution():
    v = np.ones(9) * 0.1
    v[3:6] = 1.0

    p_time = sc.DataArray(
        data=sc.array(dims=['time'], values=v),
        coords={'time': sc.linspace('time', 0.0, 8000.0, len(v), unit='us')},
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0]),
        coords={
            'wavelength': sc.array(
                dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0], unit='angstrom'
            )
        },
    )

    source = tof.Source.from_distribution(
        neutrons=123_987,
        p_time=p_time,
        p_wav=p_wav,
        pulses=2,
        frequency=sc.scalar(100.0, unit='Hz'),
    )
    assert source.data.sizes['pulse'] == 2
    assert source.data.sizes['event'] == 123_987
    bins = (sc.arange('time', 3) / source.frequency).to(unit='us')
    h = source.data.flatten(to='event').hist(time=bins)
    assert sc.allclose(
        h.data, sc.full(value=123987.0, sizes={'time': 2}, unit='counts')
    )


def test_multiple_pulses_from_neutrons():
    birth_times = sc.array(
        dims=['event'], values=[1111.0, 1567.0, 856.0, 2735.0], unit='us'
    )
    wavelengths = sc.array(
        dims=['event'], values=[1.0, 5.0, 8.0, 10.0], unit='angstrom'
    )
    source = tof.Source.from_neutrons(
        birth_times=birth_times,
        wavelengths=wavelengths,
        pulses=3,
        frequency=sc.scalar(50.0, unit='Hz'),
    )
    assert source.data.sizes['pulse'] == 3
    assert source.data.sizes['event'] == 4
    offsets = (sc.arange('pulse', 3) / source.frequency).to(unit='us')
    assert sc.allclose(source.data.coords['time'], birth_times + offsets)
    assert sc.allclose(
        source.data.coords['wavelength'],
        sc.broadcast(wavelengths, sizes={'pulse': 3, 'event': 4}),
    )


def test_source_length():
    N = 17
    source = tof.Source(facility='ess', neutrons=3124, pulses=N)
    assert len(source) == N


def test_multiple_pulses_from_distribution_no_frequency_raises():
    v = np.ones(9) * 0.1
    v[3:6] = 1.0
    p_time = sc.DataArray(
        data=sc.array(dims=['time'], values=v),
        coords={'time': sc.linspace('time', 0.0, 8.0, len(v), unit='ms')},
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0]),
        coords={
            'wavelength': sc.array(
                dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0], unit='angstrom'
            )
        },
    )
    with pytest.raises(
        ValueError, match='If pulses is greater than one, a frequency must be supplied.'
    ):
        tof.Source.from_distribution(
            neutrons=22696,
            p_time=p_time,
            p_wav=p_wav,
            pulses=3,
        )


def test_multiple_pulses_from_neutrons_no_frequency_raises():
    birth_times = sc.array(
        dims=['event'], values=[1111.0, 1567.0, 856.0, 2735.0], unit='us'
    )
    wavelengths = sc.array(
        dims=['event'], values=[1.0, 5.0, 8.0, 10.0], unit='angstrom'
    )
    with pytest.raises(
        ValueError, match='If pulses is greater than one, a frequency must be supplied.'
    ):
        tof.Source.from_neutrons(
            birth_times=birth_times,
            wavelengths=wavelengths,
            pulses=3,
        )


def test_source_repr_does_not_raise():
    assert repr(tof.Source(facility='ess', neutrons=100_000)) is not None
    assert repr(tof.Source(facility='ess', neutrons=100_000, pulses=3)) is not None



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc

import tof

from .common import (
    dummy_chopper,
    dummy_detector,
    dummy_source,
    make_chopper,
    make_source,
)

Hz = sc.Unit('Hz')
deg = sc.Unit('deg')
meter = sc.Unit('m')
ms = sc.Unit('ms')


def test_one_chopper_one_opening():
    # Make a chopper open from 10-20 ms. Assume zero phase.
    topen = 10.0 * ms
    tclose = 20.0 * ms
    chopper = make_chopper(
        topen=[topen],
        tclose=[tclose],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper',
    )
    detector = tof.Detector(distance=20 * meter, name='detector')

    # Make a pulse with 3 neutrons with one neutron going through the chopper opening
    # and the other two neutrons on either side of the opening.
    source = make_source(
        arrival_times=sc.concat(
            [0.9 * topen, 0.5 * (topen + tclose), 1.1 * tclose], dim='event'
        ),
        distance=chopper.distance,
    )

    model = tof.Model(source=source, choppers=[chopper], detectors=[detector])
    res = model.run()

    key = 'pulse:0'
    visible = res.choppers['chopper'].toas.visible.data[key]
    blocked = res.choppers['chopper'].toas.blocked.data[key]

    assert len(visible) == 1
    assert len(blocked) == 2
    assert sc.isclose(
        visible.coords['toa'][0],
        (0.5 * (topen + tclose)).to(unit='us'),
    )
    assert sc.isclose(blocked.coords['toa'][0], (0.9 * topen).to(unit='us'))
    assert sc.isclose(blocked.coords['toa'][1], (1.1 * tclose).to(unit='us'))

    det = res.detectors['detector'].toas.visible.data[key]
    assert len(det) == 1
    assert sc.isclose(
        det.coords['toa'][0],
        (
            source.data.coords['wavelength']['pulse', 0][1]
            * detector.distance
            * tof.utils.m_over_h
        ).to(unit='us'),
    )


def test_two_choppers_one_opening():
    # Make a first chopper open from 5-16 ms. Assume zero phase.
    topen = 5.0 * ms
    tclose = 16.0 * ms
    chopper1 = make_chopper(
        topen=[topen],
        tclose=[tclose],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper1',
    )

    # Make a second chopper open from 15-20 ms.
    chopper2 = make_chopper(
        topen=[15.0 * ms],
        tclose=[20.0 * ms],
        f=15.0 * Hz,
        phase=0.0 * deg,
        distance=15 * meter,
        name='chopper2',
    )

    detector = tof.Detector(distance=20 * meter, name='detector')

    # Make a pulse with 3 neutrons with 2 neutrons going through the first chopper
    # opening and only one making it through the second chopper.
    source = make_source(
        arrival_times=sc.concat(
            [1.5 * topen, 0.5 * (topen + tclose), 1.1 * tclose], dim='event'
        ),
        distance=chopper1.distance,
    )

    model = tof.Model(
        source=source, choppers=[chopper1, chopper2], detectors=[detector]
    )
    res = model.run()

    key = 'pulse:0'
    ch1_toas = res.choppers['chopper1'].toas.data
    ch2_toas = res.choppers['chopper2'].toas.data
    wavs = source.data.coords['wavelength']['pulse', 0]
    det = res.detectors['detector'].toas.visible.data[key]

    assert len(ch1_toas['visible'][key]) == 2
    assert len(ch1_toas['blocked'][key]) == 1
    assert sc.isclose(
        ch1_toas['visible'][key].coords['toa'][0], (1.5 * topen).to(unit='us')
    )
    assert sc.isclose(
        ch1_toas['visible'][key].coords['toa'][1],
        (0.5 * (topen + tclose)).to(unit='us'),
    )
    assert sc.isclose(
        ch1_toas['blocked'][key].coords['toa'][0], (1.1 * tclose).to(unit='us')
    )
    assert len(ch2_toas['visible'][key]) == 1
    # Blocks only one neutron, the other is blocked by chopper1
    assert len(ch2_toas['blocked'][key]) == 1
    assert sc.isclose(
        ch2_toas['visible'][key].coords['toa'][0],
        (wavs[1] * chopper2.distance * tof.utils.m_over_h).to(unit='us'),
    )
    assert sc.isclose(
        ch2_toas['blocked'][key].coords['toa'][0],
        (wavs[0] * chopper2.distance * tof.utils.m_over_h).to(unit='us'),
    )
    assert len(det) == 1
    assert sc.isclose(
        det.coords['toa'][0],
        (wavs[1] * detector.distance * tof.utils.m_over_h).to(unit='us'),
    )


def test_two_choppers_one_and_two_openings():
    topen = 5.0 * ms
    tclose = 16.0 * ms
    chopper1 = make_chopper(
        topen=[topen],
        tclose=[tclose],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper1',
    )

    chopper2 = make_chopper(
        topen=[9.0 * ms, 15.0 * ms],
        tclose=[12.0 * ms, 20.0 * ms],
        f=15.0 * Hz,
        phase=0.0 * deg,
        distance=15 * meter,
        name='chopper2',
    )

    detector = tof.Detector(distance=20 * meter, name='detector')

    # Make a pulse with 7 neutrons:
    # - 2 neutrons blocked by chopper1
    # - 3 neutrons blocked by chopper2
    # - 2 neutrons detected (one through each of chopper2's openings)
    source = make_source(
        arrival_times=sc.concat(
            [
                0.9 * topen,
                1.01 * topen,
                1.3 * topen,
                1.8 * topen,
                0.5 * (topen + tclose),
                0.85 * tclose,
                1.1 * tclose,
            ],
            dim='event',
        ),
        distance=chopper1.distance,
    )

    model = tof.Model(
        source=source, choppers=[chopper1, chopper2], detectors=[detector]
    )
    res = model.run()

    key = 'pulse:0'
    assert len(res.choppers['chopper1'].toas.visible.data[key]) == 5
    assert len(res.choppers['chopper1'].toas.blocked.data[key]) == 2
    assert len(res.choppers['chopper2'].toas.visible.data[key]) == 2
    assert len(res.choppers['chopper2'].toas.blocked.data[key]) == 3
    assert len(res.detectors['detector'].toas.visible.data[key]) == 2


def test_neutron_conservation():
    N = 100_000
    source = tof.Source(facility='ess', neutrons=N)

    chopper1 = make_chopper(
        topen=[5.0 * ms],
        tclose=[16.0 * ms],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper1',
    )
    chopper2 = make_chopper(
        topen=[9.0 * ms, 15.0 * ms],
        tclose=[15.0 * ms, 20.0 * ms],
        f=15.0 * Hz,
        phase=0.0 * deg,
        distance=15 * meter,
        name='chopper2',
    )

    detector = tof.Detector(distance=20 * meter, name='detector')
    model = tof.Model(
        source=source, choppers=[chopper1, chopper2], detectors=[detector]
    )
    res = model.run()

    key = 'pulse:0'
    ch1 = res.choppers['chopper1'].toas.data
    ch2 = res.choppers['chopper2'].toas.data

    assert (ch1['visible'][key].sum() + ch1['blocked'][key].sum()).value == N
    assert sc.identical(
        ch2['visible'][key].sum() + ch2['blocked'][key].sum(),
        ch1['visible'][key].sum(),
    )
    assert sc.identical(
        res.detectors['detector'].toas.data['visible'][key].sum(),
        ch2['visible'][key].sum(),
    )


def test_add_chopper_and_detector():
    chopper = dummy_chopper()
    detector = dummy_detector()
    model = tof.Model(source=dummy_source())
    model.add(chopper)
    assert 'dummy_chopper' in model.choppers
    model.add(detector)
    assert 'dummy_detector' in model.detectors


def test_add_components_with_same_name_raises():
    chopper = dummy_chopper()
    detector = dummy_detector()
    model = tof.Model(source=dummy_source())
    model.add(chopper)
    with pytest.raises(
        KeyError, match='Component with name dummy_chopper already exists'
    ):
        model.add(chopper)
    model.add(detector)
    with pytest.raises(
        KeyError, match='Component with name dummy_detector already exists'
    ):
        model.add(detector)
    detector2 = tof.Detector(distance=22 * meter, name='dummy_chopper')
    with pytest.raises(
        KeyError, match='Component with name dummy_chopper already exists'
    ):
        model.add(detector2)


def test_iter():
    chopper = dummy_chopper()
    detector = dummy_detector()
    model = tof.Model(source=dummy_source())
    model.add(chopper)
    assert 'dummy_chopper' in model.choppers
    model.add(detector)
    assert 'dummy_detector' in model.detectors


def test_remove():
    chopper = dummy_chopper()
    detector = dummy_detector()
    model = tof.Model(source=dummy_source(), choppers=[chopper], detectors=[detector])
    del model['dummy_chopper']
    assert 'dummy_chopper' not in model
    assert 'dummy_detector' in model
    del model['dummy_detector']
    assert 'dummy_detector' not in model


def test_getitem():
    chopper = dummy_chopper()
    detector = dummy_detector()
    model = tof.Model(source=dummy_source(), choppers=[chopper], detectors=[detector])
    assert model['dummy_chopper'] is chopper
    assert model['dummy_detector'] is detector
    with pytest.raises(KeyError, match='No component with name foo'):
        model['foo']


def test_input_can_be_single_component():
    chopper = dummy_chopper()
    detector = dummy_detector()
    model = tof.Model(source=dummy_source(), choppers=chopper, detectors=detector)
    assert 'dummy_chopper' in model.choppers
    assert 'dummy_detector' in model.detectors


def test_bad_input_type_raises():
    chopper = dummy_chopper()
    detector = dummy_detector()
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source(), choppers='bad chopper')
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source(), choppers=[chopper], detectors='abc')
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source(), choppers=[chopper, 'bad chopper'])
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source(), detectors=(1234, detector))
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source(), choppers=[detector])
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source(), detectors=[chopper])


def test_model_repr_does_not_raise():
    N = 10_000
    source = tof.Source(facility='ess', neutrons=N)
    chopper1 = make_chopper(
        topen=[5.0 * ms],
        tclose=[16.0 * ms],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper1',
    )
    chopper2 = make_chopper(
        topen=[9.0 * ms, 15.0 * ms],
        tclose=[15.0 * ms, 20.0 * ms],
        f=15.0 * Hz,
        phase=0.0 * deg,
        distance=15 * meter,
        name='chopper2',
    )
    detector = tof.Detector(distance=20 * meter, name='detector')
    model = tof.Model(
        source=source, choppers=[chopper1, chopper2], detectors=[detector]
    )
    assert repr(model) is not None


def test_component_distance():
    # Make a chopper open from 10-20 ms. Assume zero phase.
    topen = 10.0 * ms
    tclose = 20.0 * ms
    chopper = make_chopper(
        topen=[topen],
        tclose=[tclose],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper',
    )
    monitor = tof.Detector(distance=17 * meter, name='monitor')
    detector = tof.Detector(distance=20 * meter, name='detector')

    # Make a pulse with 3 neutrons with one neutron going through the chopper opening
    # and the other two neutrons on either side of the opening.
    source = make_source(
        arrival_times=sc.concat(
            [0.9 * topen, 0.5 * (topen + tclose), 1.1 * tclose], dim='event'
        ),
        distance=chopper.distance,
    )

    model = tof.Model(source=source, choppers=[chopper], detectors=[monitor, detector])
    res = model.run()

    assert sc.identical(res['monitor'].data.coords['distance'], monitor.distance)
    assert sc.identical(res['detector'].data.coords['distance'], detector.distance)
    assert sc.identical(res['chopper'].data.coords['distance'], chopper.distance)


def test_to_nxevent_data():
    source = tof.Source(facility='ess', neutrons=100_000)
    choppers = [
        tof.Chopper(
            frequency=14.0 * Hz,
            open=sc.array(
                dims=['cutout'],
                values=[0.0],
                unit='deg',
            ),
            close=sc.array(
                dims=['cutout'],
                values=[10.0],
                unit='deg',
            ),
            phase=90.0 * deg,
            distance=8.0 * meter,
            name="chopper",
        )
    ]
    detectors = [
        tof.Detector(distance=26.0 * meter, name='monitor'),
        tof.Detector(distance=32.0 * meter, name='detector'),
    ]
    model = tof.Model(source=source, choppers=choppers, detectors=detectors)
    res = model.run()

    # There should be 1 pulse for monitor data, and 2 pulses for detector data as it
    # wraps around the pulse period.
    for key, npulses in zip(('monitor', 'detector'), (1, 2)):
        nxevent_data = res.to_nxevent_data(key)
        assert sc.identical(res['monitor'].data.sum().data, nxevent_data.sum().data)
        assert nxevent_data.sizes['pulse'] == npulses
        assert nxevent_data.bins.concat().value.coords[
            'event_time_offset'
        ].min() >= sc.scalar(0.0, unit='us')
        assert nxevent_data.bins.concat().value.coords[
            'event_time_offset'
        ].max() <= sc.reciprocal(source.frequency).to(unit='us')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc

import tof

Hz = sc.Unit('Hz')
deg = sc.Unit('deg')
meter = sc.Unit('m')
ms = sc.Unit('ms')
two_pi = sc.constants.pi * (2.0 * sc.units.rad)


def make_chopper(topen, tclose, f, phase, distance, name):
    aopen = two_pi * sc.concat(topen, dim='cutout').to(unit='s') * f
    aclose = two_pi * sc.concat(tclose, dim='cutout').to(unit='s') * f
    return tof.Chopper(
        frequency=f,
        open=aopen,
        close=aclose,
        phase=phase,
        distance=distance,
        name=name,
    )


def make_source(arrival_times, distance, pulses=1, frequency=None):
    # Arrival times are distance * m_over_h * wavelength
    return tof.Source.from_neutrons(
        birth_times=sc.array(
            dims=['event'],
            values=[0.0] * len(arrival_times),
            unit='s',
        ),
        wavelengths=arrival_times.to(unit='s') / (distance * tof.utils.m_over_h),
        pulses=pulses,
        frequency=frequency,
    )


def dummy_chopper():
    return tof.Chopper(
        frequency=1.0 * Hz,
        open=sc.array(dims=['cutout'], values=[0.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[1.0], unit='deg'),
        phase=0.0 * deg,
        distance=1.0 * meter,
        name='dummy_chopper',
    )


def dummy_detector():
    return tof.Detector(
        distance=1.0 * meter,
        name='dummy_detector',
    )


def dummy_source():
    return tof.Source.from_neutrons(
        birth_times=sc.array(
            dims=['event'],
            values=[0.0],
            unit='s',
        ),
        wavelengths=sc.array(
            dims=['event'],
            values=[1.0],
            unit='angstrom',
        ),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from itertools import chain
from typing import Dict, List, Optional, Tuple, Union

import scipp as sc

from .chopper import Chopper
from .detector import Detector
from .result import Result
from .source import Source

ComponentType = Union[Chopper, Detector]


def _input_to_dict(
    obj: Union[
        None,
        Dict[str, ComponentType],
        List[ComponentType],
        Tuple[ComponentType, ...],
        ComponentType,
    ],
    kind: type,
):
    if isinstance(obj, (list, tuple)):
        out = {}
        for item in obj:
            out.update(_input_to_dict(item, kind=kind))
        return out
    elif isinstance(obj, kind):
        return {obj.name: obj}
    elif obj is None:
        return {}
    else:
        raise TypeError(
            "Invalid input type. Must be a Chopper or a Detector, "
            "or a list/tuple of Choppers or Detectors."
        )


class Model:
    """
    A class that represents a neutron instrument.
    It is defined by a list of choppers, a list of detectors, and a source.

    Parameters
    ----------
    choppers:
        A list of choppers.
    detectors:
        A list of detectors.
    source:
        A source of neutrons.
    """

    def __init__(
        self,
        source: Source,
        choppers: Optional[Union[Chopper, List[Chopper], Tuple[Chopper, ...]]] = None,
        detectors: Optional[
            Union[Detector, List[Detector], Tuple[Detector, ...]]
        ] = None,
    ):
        self.choppers = _input_to_dict(choppers, kind=Chopper)
        self.detectors = _input_to_dict(detectors, kind=Detector)
        self.source = source

    def add(self, component):
        """
        Add a component to the instrument.
        Component names must be unique across choppers and detectors.

        Parameters
        ----------
        component:
            A chopper or detector.
        """
        if component.name in chain(self.choppers, self.detectors):
            raise KeyError(
                f"Component with name {component.name} already exists. "
                "If you wish to replace/update an existing component, use "
                "``model.choppers['name'] = new_chopper`` or "
                "``model.detectors['name'] = new_detector``."
            )
        if isinstance(component, Chopper):
            self.choppers[component.name] = component
        elif isinstance(component, Detector):
            self.detectors[component.name] = component
        else:
            raise TypeError(
                f"Cannot add component of type {type(component)} to the model."
            )

    def remove(self, name: str):
        """
        Remove a component.

        Parameters
        ----------
        name:
            The name of the component to remove.
        """
        if name in self.choppers:
            del self.choppers[name]
        elif name in self.detectors:
            del self.detectors[name]
        else:
            raise KeyError(f"No component with name {name} was found.")

    def __iter__(self):
        return chain(self.choppers, self.detectors)

    def __getitem__(self, name) -> Union[Chopper, Detector]:
        if name not in self:
            raise KeyError(f"No component with name {name} was found.")
        return self.choppers[name] if name in self.choppers else self.detectors[name]

    def __delitem__(self, name):
        self.remove(name)

    def run(self):
        """
        Run the simulation.
        """
        components = sorted(
            chain(self.choppers.values(), self.detectors.values()),
            key=lambda c: c.distance.value,
        )

        birth_time = self.source.data.coords['time']
        speed = self.source.data.coords['speed']
        initial_mask = sc.ones(sizes=birth_time.sizes, unit=None, dtype=bool)

        result_choppers = {}
        result_detectors = {}
        time_limit = (
            birth_time + (components[-1].distance / speed).to(unit=birth_time.unit)
        ).max()
        for c in components:
            container = result_detectors if isinstance(c, Detector) else result_choppers
            container[c.name] = c.as_dict()
            container[c.name]['data'] = self.source.data.copy(deep=False)
            t = birth_time + (c.distance / speed).to(unit=birth_time.unit, copy=False)
            container[c.name]['data'].coords['toa'] = t
            container[c.name]['data'].coords['distance'] = c.distance
            # TODO: remove 'tof' coordinate once deprecation period is over
            container[c.name]['data'].coords['tof'] = t
            if isinstance(c, Detector):
                container[c.name]['visible_mask'] = initial_mask
                container[c.name]['data'].masks['blocked_by_others'] = ~initial_mask
                continue
            m = sc.zeros(sizes=t.sizes, unit=None, dtype=bool)
            to, tc = c.open_close_times(time_limit=time_limit)
            container[c.name].update({'open_times': to, 'close_times': tc})
            for i in range(len(to)):
                m |= (t > to[i]) & (t < tc[i])
            combined = initial_mask & m
            container[c.name].update(
                {'visible_mask': combined, 'blocked_mask': ~m & initial_mask}
            )
            container[c.name]['data'].masks['blocked_by_others'] = ~initial_mask
            container[c.name]['data'].masks['blocked_by_me'] = ~m & initial_mask
            initial_mask = combined

        return Result(
            source=self.source, choppers=result_choppers, detectors=result_detectors
        )

    def __repr__(self) -> str:
        out = f"Model:\n  Source: {self.source}\n  Choppers:\n"
        for name, ch in self.choppers.items():
            out += f"    {name}: {ch}\n"
        out += "  Detectors:\n"
        for name, det in self.detectors.items():
            out += f"    {name}: {det}\n"
        return out

    def __str__(self) -> str:
        return self.__repr__()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from .chopper import AntiClockwise, Chopper, ChopperReading, Clockwise
from .detector import Detector, DetectorReading
from .facilities import library as facilities
from .model import Model
from .reading import ComponentReading, ReadingData, ReadingField
from .result import Result
from .source import Source, SourceParameters

__all__ = [
    'AntiClockwise',
    'Chopper',
    'ChopperReading',
    'Clockwise',
    'ComponentReading',
    'Detector',
    'DetectorReading',
    'facilities',
    'Model',
    'ReadingData',
    'ReadingField',
    'Result',
    'Source',
    'SourceParameters',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from dataclasses import dataclass
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import plopp as pp
import scipp as sc
from scipp.scipy.interpolate import interp1d

from .facilities import library as facilities
from .utils import Plot, wavelength_to_speed

TIME_UNIT = "us"
WAV_UNIT = "angstrom"


def _default_frequency(frequency: Union[None, sc.Variable], pulses: int) -> sc.Variable:
    if frequency is None:
        if pulses > 1:
            raise ValueError(
                "If pulses is greater than one, a frequency must be supplied."
            )
        frequency = 1.0 * sc.Unit("Hz")
    return frequency


def _convert_coord(da: sc.DataArray, unit: str, coord: str) -> sc.DataArray:
    out = da.copy(deep=False)
    out.coords[coord] = out.coords[coord].to(dtype=float, unit=unit)
    return out


def _make_pulses(
    neutrons: int,
    frequency: sc.Variable,
    pulses: int,
    p_time: sc.DataArray,
    p_wav: sc.DataArray,
    sampling: int,
    wmin: Optional[sc.Variable] = None,
    wmax: Optional[sc.Variable] = None,
):
    """
    Create pulses from time a wavelength probability distributions.
    The distributions should be supplied as DataArrays where the coordinates
    are the values of the distribution, and the values are the probability.
    Note that the time and wavelength distributions are independent. A neutron with
    a randomly selected birth time from ``p_time`` can adopt any wavelength in
    ``p_wav`` (in other words, the two distributions are simply broadcast into a
    square 2D parameter space).

    Parameters
    ----------
    neutrons:
        Number of neutrons per pulse.
    frequency:
        Pulse frequency.
    pulses:
        Number of pulses.
    p_time:
        Time probability distribution for a single pulse.
    p_wav:
        Wavelength probability distribution for a single pulse.
    sampling:
        Number of points used to sample the probability distributions.
    wmin:
        Minimum neutron wavelength.
    wmax:
        Maximum neutron wavelength.
    """
    t_dim = "time"
    w_dim = "wavelength"

    p_time = _convert_coord(p_time, unit=TIME_UNIT, coord=t_dim)
    p_wav = _convert_coord(p_wav, unit=WAV_UNIT, coord=w_dim)
    sampling = int(sampling)

    tmin = p_time.coords[t_dim].min()
    tmax = p_time.coords[t_dim].max()
    if wmin is None:
        wmin = p_wav.coords[w_dim].min()
    if wmax is None:
        wmax = p_wav.coords[w_dim].max()

    time_interpolator = interp1d(p_time, dim=t_dim, fill_value="extrapolate")
    wav_interpolator = interp1d(p_wav, dim=w_dim, fill_value="extrapolate")
    x_time = sc.linspace(
        dim=t_dim,
        start=tmin.value,
        stop=tmax.value,
        num=sampling,
        unit=TIME_UNIT,
    )
    x_wav = sc.linspace(
        dim=w_dim,
        start=wmin.value,
        stop=wmax.value,
        num=sampling,
        unit=WAV_UNIT,
    )
    p_time = time_interpolator(x_time)
    p_time /= p_time.data.sum()
    p_wav = wav_interpolator(x_wav)
    p_wav /= p_wav.data.sum()

    # In the following, random.choice only allows to select from the values listed
    # in the coordinate of the probability distribution arrays. This leads to data
    # grouped into spikes and empty in between because the sampling resolution used
    # in the linear interpolation above is usually kept low for performance.
    # To make the distribution more uniform, we add some random noise to the chosen
    # values, which effectively fills in the gaps between the spikes.
    # Scipy has some methods to sample from a continuous distribution, but they are
    # prohibitively slow.
    # See https://docs.scipy.org/doc/scipy/tutorial/stats/sampling.html for more
    # information.
    dt = 0.5 * (tmax - tmin).value / (p_time.sizes[t_dim] - 1)
    dw = 0.5 * (wmax - wmin).value / (p_wav.sizes[w_dim] - 1)

    # Because of the added noise, some values end up being outside the specified range
    # for the birth times and wavelengths. Using naive clipping leads to pile-up on the
    # edges of the range. To avoid this, we remove the outliers and resample until we
    # have the desired number of neutrons.
    n = 0
    times = []
    wavs = []
    ntot = pulses * neutrons
    while n < ntot:
        size = ntot - n
        t = np.random.choice(
            p_time.coords[t_dim].values, size=size, p=p_time.values
        ) + np.random.normal(scale=dt, size=size)
        w = np.random.choice(
            p_wav.coords[w_dim].values, size=size, p=p_wav.values
        ) + np.random.normal(scale=dw, size=size)
        mask = (
            (t >= tmin.value)
            & (t <= tmax.value)
            & (w >= wmin.value)
            & (w <= wmax.value)
        )
        times.append(t[mask])
        wavs.append(w[mask])
        n += mask.sum()

    dim = "event"
    birth_time = sc.array(
        dims=[dim],
        values=np.concatenate(times),
        unit=TIME_UNIT,
    ).fold(dim=dim, sizes={"pulse": pulses, dim: neutrons}) + (
        sc.arange("pulse", pulses) / frequency
    ).to(
        unit=TIME_UNIT, copy=False
    )

    wavelength = sc.array(
        dims=[dim],
        values=np.concatenate(wavs),
        unit=WAV_UNIT,
    ).fold(dim=dim, sizes={"pulse": pulses, dim: neutrons})
    speed = wavelength_to_speed(wavelength)
    return {
        "time": birth_time,
        "wavelength": wavelength,
        "speed": speed,
    }


class Source:
    """
    A class that represents a source of neutrons.
    It is defined by the number of neutrons, a wavelength range, and a time range.
    The default way of creating a pulse is to supply the name of a facility
    (e.g. ``'ess'``) and the number of neutrons. This will create a pulse with the
    default time and wavelength ranges for that facility.

    Parameters
    ----------
    facility:
        Name of a pre-defined pulse shape from a neutron facility.
    neutrons:
        Number of neutrons per pulse.
    pulses:
        Number of pulses.
    sampling:
        Number of points used to interpolate the probability distributions.
    wmin:
        Minimum neutron wavelength.
    wmax:
        Maximum neutron wavelength.
    """

    def __init__(
        self,
        facility: str,
        neutrons: int = 1_000_000,
        pulses: int = 1,
        sampling: int = 1000,
        wmin: Optional[sc.Variable] = None,
        wmax: Optional[sc.Variable] = None,
    ):
        self.facility = facility
        self.neutrons = int(neutrons)
        self.pulses = int(pulses)
        self.data = None

        if facility is not None:
            facility_params = facilities[self.facility]
            self.frequency = facility_params.frequency
            pulse_params = _make_pulses(
                neutrons=self.neutrons,
                p_time=facility_params.time,
                p_wav=facility_params.wavelength,
                sampling=sampling,
                frequency=self.frequency,
                pulses=pulses,
                wmin=wmin,
                wmax=wmax,
            )
            self.data = sc.DataArray(
                data=sc.ones(sizes=pulse_params["time"].sizes, unit="counts"),
                coords={
                    "time": pulse_params["time"],
                    "wavelength": pulse_params["wavelength"],
                    "speed": pulse_params["speed"],
                    "id": sc.arange("event", pulse_params["time"].size, unit=None).fold(
                        "event", sizes=pulse_params["time"].sizes
                    ),
                },
            )

    @classmethod
    def from_neutrons(
        cls,
        birth_times: sc.Variable,
        wavelengths: sc.Variable,
        frequency: Optional[sc.Variable] = None,
        pulses: int = 1,
    ):
        """
        Create source pulses from a list of neutrons.
        Both ``birth times`` and ``wavelengths`` should be one-dimensional and have the
        same length. They represent the neutrons inside one pulse. If ``pulses`` is
        greater than one, the neutrons will be repeated ``pulses`` times.

        Parameters
        ----------
        birth_times:
            Birth times of neutrons in the pulse.
        wavelengths:
            Wavelengths of neutrons in the pulse.
        frequency:
            Frequency of the pulse.
        pulses:
            Number of pulses.
        """
        source = cls(facility=None, neutrons=len(birth_times), pulses=pulses)
        source.frequency = _default_frequency(frequency, pulses)

        birth_times = (sc.arange("pulse", pulses) / source.frequency).to(
            unit=TIME_UNIT, copy=False
        ) + birth_times.to(unit=TIME_UNIT, copy=False)
        wavelengths = sc.broadcast(
            wavelengths.to(unit=WAV_UNIT, copy=False), sizes=birth_times.sizes
        )

        source.data = sc.DataArray(
            data=sc.ones(sizes=birth_times.sizes, unit="counts"),
            coords={
                "time": birth_times,
                "wavelength": wavelengths,
                "speed": wavelength_to_speed(wavelengths).to(unit="m/s", copy=False),
                "id": sc.arange("event", birth_times.size, unit=None).fold(
                    "event", sizes=birth_times.sizes
                ),
            },
        )

        return source

    @classmethod
    def from_distribution(
        cls,
        p_time: sc.DataArray,
        p_wav: sc.DataArray,
        neutrons: int = 1_000_000,
        pulses: int = 1,
        frequency: Optional[sc.Variable] = None,
        sampling: Optional[int] = 1000,
    ):
        """
        Create source pulses from time a wavelength probability distributions.
        The distributions should be supplied as DataArrays where the coordinates
        are the values of the distribution, and the values are the probability.
        Note that the time and wavelength distributions are independent. A neutron with
        a randomly selected birth time from ``p_time`` can adopt any wavelength in
        ``p_wav`` (in other words, the two distributions are simply broadcast into a
        square 2D parameter space).

        Parameters
        ----------
        p_time:
            Time probability distribution.
        p_wav:
            Wavelength probability distribution.
        neutrons:
            Number of neutrons in the pulse.
        pulses:
            Number of pulses.
        frequency:
            Frequency of the pulse.
        sampling:
            Number of points used to interpolate the probability distributions.
        """

        source = cls(facility=None, neutrons=neutrons, pulses=pulses)
        source.frequency = _default_frequency(frequency, pulses)
        pulse_params = _make_pulses(
            neutrons=neutrons,
            p_time=p_time,
            p_wav=p_wav,
            frequency=source.frequency,
            pulses=pulses,
            sampling=sampling,
        )
        source.data = sc.DataArray(
            data=sc.ones(sizes=pulse_params["time"].sizes, unit="counts"),
            coords={
                "time": pulse_params["time"],
                "wavelength": pulse_params["wavelength"],
                "speed": pulse_params["speed"],
                "id": sc.arange("event", pulse_params["time"].size, unit=None).fold(
                    "event", sizes=pulse_params["time"].sizes
                ),
            },
        )
        return source

    def __len__(self) -> int:
        return self.data.sizes["pulse"]

    def plot(self, bins: int = 300) -> tuple:
        """
        Plot the pulses of the source.

        Parameters
        ----------
        bins:
            Number of bins to use for histogramming the neutrons.
        """
        fig, ax = plt.subplots(1, 2)
        dim = (set(self.data.dims) - {"pulse"}).pop()
        collapsed = sc.collapse(self.data, keep=dim)
        pp.plot(
            {k: da.hist(time=bins) for k, da in collapsed.items()},
            ax=ax[0],
        )
        pp.plot(
            {k: da.hist(wavelength=bins) for k, da in collapsed.items()},
            ax=ax[1],
        )
        fig.set_size_inches(10, 4)
        fig.tight_layout()
        return Plot(fig=fig, ax=ax)

    def as_readonly(self):
        return SourceParameters(
            data=self.data,
            facility=self.facility,
            neutrons=self.neutrons,
            frequency=self.frequency,
            pulses=self.pulses,
        )

    def __repr__(self) -> str:
        return (
            f"Source:\n"
            f"  pulses={self.pulses}, neutrons per pulse={self.neutrons}\n"
            f"  frequency={self.frequency:c}\n  facility='{self.facility}'"
        )


@dataclass(frozen=True)
class SourceParameters:
    """
    Read-only container for the parameters of a source.
    """

    data: sc.DataArray
    facility: Optional[str]
    neutrons: int
    frequency: sc.Variable
    pulses: int



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from .ess_pulse import pulse as esspulse

library = {'ess': esspulse}

__all__ = ['library']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from dataclasses import dataclass

import matplotlib.pyplot as plt
import scipp as sc
import scipp.constants as const

m_over_h = const.m_n / const.h
two_pi = sc.constants.pi * (2.0 * sc.units.rad)


def speed_to_wavelength(x: sc.Variable, unit: str = 'angstrom') -> sc.Variable:
    """
    Convert neutron speeds to wavelengths.

    Parameters
    ----------
    x:
        Input speeds.
    unit:
        The unit of the output wavelengths.
    """
    return (1.0 / (m_over_h * x)).to(unit=unit)


def wavelength_to_speed(x: sc.Variable, unit: str = 'm/s') -> sc.Variable:
    """
    Convert neutron wavelengths to speeds.

    Parameters
    ----------
    x:
        Input wavelengths.
    unit:
        The unit of the output speeds.
    """
    return (1.0 / (m_over_h * x)).to(unit=unit)


def speed_to_energy(x: sc.Variable, unit='meV') -> sc.Variable:
    """
    Convert neutron speeds to energies.

    Parameters
    ----------
    x:
        Input speeds.
    unit:
        The unit of the output energies.
    """
    return (const.m_n * x * x).to(unit=unit)


def energy_to_speed(x: sc.Variable, unit='m/s') -> sc.Variable:
    """
    Convert neutron energies to speeds.

    Parameters
    ----------
    x:
        Input energies.
    unit:
        The unit of the output speeds.
    """
    return sc.sqrt(x / const.m_n).to(unit=unit)


@dataclass(frozen=True)
class FacilityPulse:
    time: sc.DataArray
    wavelength: sc.DataArray
    frequency: sc.Variable


@dataclass
class Plot:
    ax: plt.Axes
    fig: plt.Figure
